#    Copyright 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_messaging import exceptions


class ExchangeNotFoundException(exceptions.MessageDeliveryFailure):
    """Is raised if specified exchange is not found in RabbitMQ."""
    pass


class MessageRejectedException(exceptions.MessageDeliveryFailure):
    """Is raised if message which you are trying to send was nacked by RabbitMQ
    it may happen if RabbitMQ is not able to process message
    """
    pass


class RoutingException(exceptions.MessageDeliveryFailure):
    """Is raised if message can not be delivered to any queue. Usually it means
    that any queue is not binded to given exchange with given routing key.
    Raised if 'mandatory' flag specified only
    """
    pass


class ConnectionException(exceptions.MessagingException):
    """Is raised if some operation can not be performed due to connectivity
    problem
    """
    pass


class TimeoutConnectionException(ConnectionException):
    """Is raised if socket timeout was expired during network interaction"""
    pass


class EstablishConnectionException(ConnectionException):
    """Is raised if we have some problem during establishing connection
    procedure
    """
    pass


class HostConnectionNotAllowedException(EstablishConnectionException):
    """Is raised in case of try to establish connection to temporary
    not allowed host (because of reconnection policy for example)
    """
    pass


class UnsupportedDriverVersion(exceptions.MessagingException):
    """Is raised when message is received but was sent by different,
    not supported driver version
    """
    pass



#    Copyright 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import logging
import threading

import six

from oslo_messaging._drivers.zmq_driver import zmq_async
from oslo_messaging._drivers.zmq_driver import zmq_names
from oslo_messaging._drivers.zmq_driver import zmq_version

LOG = logging.getLogger(__name__)

zmq = zmq_async.import_zmq()


@six.add_metaclass(abc.ABCMeta)
class SenderBase(object):
    """Base request/response sending interface."""

    def __init__(self, conf, async=False):
        self.conf = conf
        self.async = async
        self._lock = threading.Lock()
        self._send_versions = zmq_version.get_method_versions(self, 'send')

    def _get_send_version(self, version):
        send_version = self._send_versions.get(version)
        if send_version is None:
            raise zmq_version.UnsupportedMessageVersionError(version)
        return send_version

    @abc.abstractmethod
    def send(self, socket, message):
        """Send a message via a socket in a thread-safe manner."""


class RequestSenderBase(SenderBase):
    pass


class AckSenderBase(SenderBase):
    pass


class ReplySenderBase(SenderBase):
    pass


class RequestSenderProxy(RequestSenderBase):

    def send(self, socket, request):
        assert request.msg_type in zmq_names.REQUEST_TYPES, "Request expected!"

        send_version = self._get_send_version(request.message_version)

        with self._lock:
            send_version(socket, request)

        LOG.debug("->[proxy:%(addr)s] Sending %(msg_type)s message "
                  "%(msg_id)s to target %(target)s (v%(msg_version)s)",
                  {"addr": list(socket.connections),
                   "msg_type": zmq_names.message_type_str(request.msg_type),
                   "msg_id": request.message_id,
                   "target": request.target,
                   "msg_version": request.message_version})

    def _send_v_1_0(self, socket, request):
        socket.send(b'', zmq.SNDMORE)
        socket.send_string('1.0', zmq.SNDMORE)
        socket.send(six.b(str(request.msg_type)), zmq.SNDMORE)
        socket.send(request.routing_key, zmq.SNDMORE)
        socket.send_string(request.message_id, zmq.SNDMORE)
        socket.send_dumped([request.context, request.message])


class AckSenderProxy(AckSenderBase):

    def send(self, socket, ack):
        assert ack.msg_type == zmq_names.ACK_TYPE, "Ack expected!"

        send_version = self._get_send_version(ack.message_version)

        with self._lock:
            send_version(socket, ack)

        LOG.debug("->[proxy:%(addr)s] Sending %(msg_type)s for %(msg_id)s "
                  "(v%(msg_version)s)",
                  {"addr": list(socket.connections),
                   "msg_type": zmq_names.message_type_str(ack.msg_type),
                   "msg_id": ack.message_id,
                   "msg_version": ack.message_version})

    def _send_v_1_0(self, socket, ack):
        socket.send(b'', zmq.SNDMORE)
        socket.send_string('1.0', zmq.SNDMORE)
        socket.send(six.b(str(ack.msg_type)), zmq.SNDMORE)
        socket.send(ack.reply_id, zmq.SNDMORE)
        socket.send_string(ack.message_id)


class ReplySenderProxy(ReplySenderBase):

    def send(self, socket, reply):
        assert reply.msg_type == zmq_names.REPLY_TYPE, "Reply expected!"

        send_version = self._get_send_version(reply.message_version)

        with self._lock:
            send_version(socket, reply)

        LOG.debug("->[proxy:%(addr)s] Sending %(msg_type)s for %(msg_id)s "
                  "(v%(msg_version)s)",
                  {"addr": list(socket.connections),
                   "msg_type": zmq_names.message_type_str(reply.msg_type),
                   "msg_id": reply.message_id,
                   "msg_version": reply.message_version})

    def _send_v_1_0(self, socket, reply):
        socket.send(b'', zmq.SNDMORE)
        socket.send_string('1.0', zmq.SNDMORE)
        socket.send(six.b(str(reply.msg_type)), zmq.SNDMORE)
        socket.send(reply.reply_id, zmq.SNDMORE)
        socket.send_string(reply.message_id, zmq.SNDMORE)
        socket.send_dumped([reply.reply_body, reply.failure])


class RequestSenderDirect(RequestSenderBase):

    def send(self, socket, request):
        assert request.msg_type in zmq_names.REQUEST_TYPES, "Request expected!"

        send_version = self._get_send_version(request.message_version)

        with self._lock:
            send_version(socket, request)

        LOG.debug("Sending %(msg_type)s message %(msg_id)s to "
                  "target %(target)s (v%(msg_version)s)",
                  {"msg_type": zmq_names.message_type_str(request.msg_type),
                   "msg_id": request.message_id,
                   "target": request.target,
                   "msg_version": request.message_version})

    def _send_v_1_0(self, socket, request):
        flags = zmq.NOBLOCK if self.async else 0
        socket.send(b'', zmq.SNDMORE | flags)
        socket.send_string('1.0', zmq.SNDMORE | flags)
        socket.send(six.b(str(request.msg_type)), zmq.SNDMORE | flags)
        socket.send_string(request.message_id, zmq.SNDMORE | flags)
        socket.send_dumped([request.context, request.message], flags)


class AckSenderDirect(AckSenderBase):

    def send(self, socket, ack):
        assert ack.msg_type == zmq_names.ACK_TYPE, "Ack expected!"

        send_version = self._get_send_version(ack.message_version)

        with self._lock:
            send_version(socket, ack)

        LOG.debug("Sending %(msg_type)s for %(msg_id)s (v%(msg_version)s)",
                  {"msg_type": zmq_names.message_type_str(ack.msg_type),
                   "msg_id": ack.message_id,
                   "msg_version": ack.message_version})

    def _send_v_1_0(self, socket, ack):
        raise NotImplementedError()


class ReplySenderDirect(ReplySenderBase):

    def send(self, socket, reply):
        assert reply.msg_type == zmq_names.REPLY_TYPE, "Reply expected!"

        send_version = self._get_send_version(reply.message_version)

        with self._lock:
            send_version(socket, reply)

        LOG.debug("Sending %(msg_type)s for %(msg_id)s (v%(msg_version)s)",
                  {"msg_type": zmq_names.message_type_str(reply.msg_type),
                   "msg_id": reply.message_id,
                   "msg_version": reply.message_version})

    def _send_v_1_0(self, socket, reply):
        socket.send(reply.reply_id, zmq.SNDMORE)
        socket.send(b'', zmq.SNDMORE)
        socket.send_string('1.0', zmq.SNDMORE)
        socket.send(six.b(str(reply.msg_type)), zmq.SNDMORE)
        socket.send_string(reply.message_id, zmq.SNDMORE)
        socket.send_dumped([reply.reply_body, reply.failure])



#    Copyright 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import functools
import logging
import random
import time

from oslo_config import cfg
from oslo_utils import importutils
import six
import tenacity

from oslo_messaging._drivers.zmq_driver.matchmaker import zmq_matchmaker_base
from oslo_messaging._drivers.zmq_driver import zmq_address
from oslo_messaging._drivers.zmq_driver import zmq_updater
from oslo_messaging._i18n import _LE, _LI, _LW

redis = importutils.try_import('redis')
redis_sentinel = importutils.try_import('redis.sentinel')

LOG = logging.getLogger(__name__)


matchmaker_redis_opts = [
    cfg.StrOpt('host',
               default='127.0.0.1',
               deprecated_for_removal=True,
               deprecated_reason="Replaced by [DEFAULT]/transport_url",
               help='Host to locate redis.'),
    cfg.PortOpt('port',
                default=6379,
                deprecated_for_removal=True,
                deprecated_reason="Replaced by [DEFAULT]/transport_url",
                help='Use this port to connect to redis host.'),
    cfg.StrOpt('password',
               default='',
               secret=True,
               deprecated_for_removal=True,
               deprecated_reason="Replaced by [DEFAULT]/transport_url",
               help='Password for Redis server (optional).'),
    cfg.ListOpt('sentinel_hosts',
                default=[],
                deprecated_for_removal=True,
                deprecated_reason="Replaced by [DEFAULT]/transport_url",
                help='List of Redis Sentinel hosts (fault tolerance mode), '
                     'e.g., [host:port, host1:port ... ]'),
    cfg.StrOpt('sentinel_group_name',
               default='oslo-messaging-zeromq',
               help='Redis replica set name.'),
    cfg.IntOpt('wait_timeout',
               default=2000,
               help='Time in ms to wait between connection attempts.'),
    cfg.IntOpt('check_timeout',
               default=20000,
               help='Time in ms to wait before the transaction is killed.'),
    cfg.IntOpt('socket_timeout',
               default=10000,
               help='Timeout in ms on blocking socket operations.'),
]

_PUBLISHERS_KEY = "PUBLISHERS"
_ROUTERS_KEY = "ROUTERS"


def write_to_redis_connection_warn(func):
    @functools.wraps(func)
    def func_wrapper(self, *args, **kwargs):
        # try to perform a write operation to all available hosts
        success = False
        for redis_instance in self._redis_instances:
            if not redis_instance._is_available:
                continue
            try:
                func(self, redis_instance, *args, **kwargs)
                success = True
            except redis.ConnectionError:
                LOG.warning(_LW("Redis host %s is not available now."),
                            redis_instance._address)
                redis_instance._is_available = False
                redis_instance._ready_from = float("inf")
        if not success:
            raise zmq_matchmaker_base.MatchmakerUnavailable()
    return func_wrapper


def read_from_redis_connection_warn(func):
    @functools.wraps(func)
    def func_wrapper(self, *args, **kwargs):
        # try to perform a read operation from any available and ready host
        for redis_instance in self._redis_instances:
            if not redis_instance._is_available \
                    or redis_instance._ready_from > time.time():
                continue
            try:
                return func(self, redis_instance, *args, **kwargs)
            except redis.ConnectionError:
                LOG.warning(_LW("Redis host %s is not available now."),
                            redis_instance._address)
                redis_instance._is_available = False
                redis_instance._ready_from = float("inf")
        raise zmq_matchmaker_base.MatchmakerUnavailable()
    return func_wrapper


def no_reraise(func):
    def func_wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except zmq_matchmaker_base.MatchmakerUnavailable:
            pass
    return func_wrapper


def empty_list_on_error(func):
    def func_wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except zmq_matchmaker_base.MatchmakerUnavailable:
            return []
    return func_wrapper


def is_empty(hosts):
    return not hosts


@six.add_metaclass(abc.ABCMeta)
class MatchmakerRedisBase(zmq_matchmaker_base.MatchmakerBase):

    def __init__(self, conf, *args, **kwargs):
        if redis is None:
            raise ImportError(_LE("Redis package is not available!"))

        super(MatchmakerRedisBase, self).__init__(conf, *args, **kwargs)

        self.conf.register_opts(matchmaker_redis_opts, "matchmaker_redis")

    @abc.abstractmethod
    def _sadd(self, key, value, expire):
        pass

    @abc.abstractmethod
    def _srem(self, key, value):
        pass

    @abc.abstractmethod
    def _smembers(self, key):
        pass

    @abc.abstractmethod
    def _ttl(self, key):
        pass

    @no_reraise
    def register_publisher(self, hostname, expire=-1):
        hostname = ','.join(hostname)
        self._sadd(_PUBLISHERS_KEY, hostname, expire)
        self._sadd(hostname, ' ', expire)

    @no_reraise
    def unregister_publisher(self, hostname):
        hostname = ','.join(hostname)
        self._srem(_PUBLISHERS_KEY, hostname)
        self._srem(hostname, ' ')

    @empty_list_on_error
    def get_publishers(self):
        return [tuple(hostname.split(',')) for hostname
                in self._smembers(_PUBLISHERS_KEY)]

    @no_reraise
    def register_router(self, hostname, expire=-1):
        self._sadd(_ROUTERS_KEY, hostname, expire)
        self._sadd(hostname, ' ', expire)

    @no_reraise
    def unregister_router(self, hostname):
        self._srem(_ROUTERS_KEY, hostname)
        self._srem(hostname, ' ')

    @empty_list_on_error
    def get_routers(self):
        return self._smembers(_ROUTERS_KEY)

    def get_hosts_by_key(self, key):
        return self._smembers(key)

    def register(self, target, hostname, listener_type, expire=-1):
        if target.server:
            key = zmq_address.target_to_key(target, listener_type)
            self._sadd(key, hostname, expire)
            self._sadd(hostname, ' ', expire)

        key = zmq_address.prefix_str(target.topic, listener_type)
        self._sadd(key, hostname, expire)
        self._sadd(hostname, ' ', expire)

    @no_reraise
    def unregister(self, target, hostname, listener_type):
        if target.server:
            key = zmq_address.target_to_key(target, listener_type)
            self._srem(key, hostname)
            self._srem(hostname, ' ')

        key = zmq_address.prefix_str(target.topic, listener_type)
        self._srem(key, hostname)
        self._srem(hostname, ' ')

    def get_hosts(self, target, listener_type):
        hosts = []

        if target.server:
            key = zmq_address.target_to_key(target, listener_type)
            hosts.extend(self._smembers(key))
        else:
            key = zmq_address.prefix_str(target.topic, listener_type)
            hosts.extend(self._smembers(key))

        LOG.debug("[Redis] get_hosts for target %(target)s: %(hosts)s",
                  {"target": target, "hosts": hosts})

        return hosts

    def get_hosts_retry(self, target, listener_type):
        return self._retry_method(target, listener_type, self.get_hosts)

    def get_hosts_fanout(self, target, listener_type):
        key = zmq_address.target_to_key(target, listener_type)
        hosts = list(self._smembers(key))

        LOG.debug("[Redis] get_hosts_fanout for target %(target)s: %(hosts)s",
                  {"target": target, "hosts": hosts})

        return hosts

    def get_hosts_fanout_retry(self, target, listener_type):
        return self._retry_method(target, listener_type, self.get_hosts_fanout)

    def _retry_method(self, target, listener_type, method):
        wait_timeout = self.conf.matchmaker_redis.wait_timeout / 1000.
        check_timeout = self.conf.matchmaker_redis.check_timeout / 1000.

        @tenacity.retry(retry=tenacity.retry_if_result(is_empty),
                        wait=tenacity.wait_fixed(wait_timeout),
                        stop=tenacity.stop_after_delay(check_timeout))
        def _get_hosts_retry(target, listener_type):
            return method(target, listener_type)

        return _get_hosts_retry(target, listener_type)


class MatchmakerRedis(MatchmakerRedisBase):

    def __init__(self, conf, *args, **kwargs):
        super(MatchmakerRedis, self).__init__(conf, *args, **kwargs)

        self._redis_hosts = self._extract_redis_hosts()

        self._redis_instances = [
            redis.StrictRedis(host=redis_host["host"],
                              port=redis_host["port"],
                              password=redis_host["password"])
            for redis_host in self._redis_hosts
        ]

        for redis_host, redis_instance \
                in six.moves.zip(self._redis_hosts, self._redis_instances):
            address = "{host}:{port}".format(host=redis_host["host"],
                                             port=redis_host["port"])
            redis_instance._address = address
            is_available = self._check_availability(redis_instance)
            if is_available:
                redis_instance._is_available = True
                redis_instance._ready_from = time.time()
            else:
                LOG.warning(_LW("Redis host %s is not available now."),
                            address)
                redis_instance._is_available = False
                redis_instance._ready_from = float("inf")

        # NOTE(gdavoian): store instances in a random order
        # (for the sake of load balancing)
        random.shuffle(self._redis_instances)

        self._availability_updater = \
            MatchmakerRedisAvailabilityUpdater(self.conf, self)

    def _extract_redis_hosts(self):
        if self.url and self.url.hosts:
            return [{"host": redis_host.hostname,
                     "port": redis_host.port,
                     "password": redis_host.password}
                    for redis_host in self.url.hosts]
        else:
            # FIXME(gdavoian): remove the code below along with the
            # corresponding deprecated options in the next release
            return [{"host": self.conf.matchmaker_redis.host,
                     "port": self.conf.matchmaker_redis.port,
                     "password": self.conf.matchmaker_redis.password}]

    @staticmethod
    def _check_availability(redis_instance):
        try:
            redis_instance.ping()
            return True
        except redis.ConnectionError:
            return False

    @write_to_redis_connection_warn
    def _sadd(self, redis_instance, key, value, expire):
        redis_instance.sadd(key, value)
        if expire > 0:
            redis_instance.expire(key, expire)

    @write_to_redis_connection_warn
    def _srem(self, redis_instance, key, value):
        redis_instance.srem(key, value)

    @read_from_redis_connection_warn
    def _ttl(self, redis_instance, key):
        # NOTE(ozamiatin): If the specialized key doesn't exist,
        # TTL fuction would return -2. If key exists,
        # but doesn't have expiration associated,
        # TTL func would return -1. For more information,
        # please visit http://redis.io/commands/ttl
        return redis_instance.ttl(key)

    @read_from_redis_connection_warn
    def _smembers(self, redis_instance, key):
        hosts = redis_instance.smembers(key)
        return [host for host in hosts if redis_instance.ttl(host) >= -1]


class MatchmakerRedisAvailabilityUpdater(zmq_updater.UpdaterBase):

    _MIN_SLEEP_FOR = 10

    def __init__(self, conf, matchmaker):
        super(MatchmakerRedisAvailabilityUpdater, self).__init__(
            conf, matchmaker, self._update_availability,
            sleep_for=conf.oslo_messaging_zmq.zmq_target_update
        )

    def _update_availability(self):
        fraction_of_available_instances = 0
        for redis_instance in self.matchmaker._redis_instances:
            if not redis_instance._is_available:
                is_available = \
                    self.matchmaker._check_availability(redis_instance)
                if is_available:
                    LOG.info(_LI("Redis host %s is available again."),
                             redis_instance._address)
                    fraction_of_available_instances += 1
                    # NOTE(gdavoian): mark an instance as available for
                    # writing to, but wait until all services register
                    # themselves in it for making the instance ready for
                    # reading from
                    redis_instance._is_available = True
                    redis_instance._ready_from = time.time() + \
                        self.conf.oslo_messaging_zmq.zmq_target_expire
            else:
                fraction_of_available_instances += 1
        fraction_of_available_instances /= \
            float(len(self.matchmaker._redis_instances))
        # NOTE(gdavoian): make the sleep time proportional to the number of
        # currently available instances
        self._sleep_for = max(self.conf.oslo_messaging_zmq.zmq_target_update *
                              fraction_of_available_instances,
                              self._MIN_SLEEP_FOR)


class MatchmakerSentinel(MatchmakerRedisBase):

    def __init__(self, conf, *args, **kwargs):
        super(MatchmakerSentinel, self).__init__(conf, *args, **kwargs)
        socket_timeout = self.conf.matchmaker_redis.socket_timeout / 1000.
        self._sentinel_hosts, self._password, self._master_group = \
            self._extract_sentinel_hosts()
        self._sentinel = redis_sentinel.Sentinel(
            sentinels=self._sentinel_hosts,
            socket_timeout=socket_timeout,
            password=self._password)
        self._slave = self._master = None

    @property
    def _redis_master(self):
        try:
            if not self._master:
                self._master = self._sentinel.master_for(self._master_group)
            return self._master
        except redis_sentinel.MasterNotFoundError:
            raise zmq_matchmaker_base.MatchmakerUnavailable()

    @property
    def _redis_slave(self):
        try:
            if not self._slave:
                self._slave = self._sentinel.slave_for(self._master_group)
        except redis_sentinel.SlaveNotFoundError:
            # use the master as slave (temporary)
            return self._redis_master
        return self._slave

    def _extract_sentinel_hosts(self):

        sentinels = []
        master_group = self.conf.matchmaker_redis.sentinel_group_name
        master_password = None

        if self.url and self.url.hosts:
            for host in self.url.hosts:
                target = host.hostname, host.port
                if host.password:
                    master_password = host.password
                sentinels.append(target)
            if self.url.virtual_host:
                # url://:pass@sentinel_a,:pass@sentinel_b/master_group_name
                master_group = self.url.virtual_host
        elif self.conf.matchmaker_redis.sentinel_hosts:
            s = self.conf.matchmaker_redis.sentinel_hosts
            sentinels.extend([tuple(target.split(":")) for target in s])
            master_password = self.conf.matchmaker_redis.password

        return sentinels, master_password, master_group

    def _sadd(self, key, value, expire):
        self._redis_master.sadd(key, value)
        if expire > 0:
            self._redis_master.expire(key, expire)

    def _srem(self, key, value):
        self._redis_master.srem(key, value)

    def _smembers(self, key):
        hosts = self._redis_slave.smembers(key)
        return [host for host in hosts if self._ttl(host) >= -1]

    def _ttl(self, key):
        return self._redis_slave.ttl(key)



#    Copyright 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import logging
import os
import threading

import futurist
from pika.adapters import select_connection
from pika import exceptions as pika_exceptions
from pika import spec as pika_spec

from oslo_utils import eventletutils

current_thread = eventletutils.fetch_current_thread_functor()

LOG = logging.getLogger(__name__)


class ThreadSafePikaConnection(object):
    def __init__(self, parameters=None,
                 _impl_class=select_connection.SelectConnection):
        self.params = parameters
        self._connection_lock = threading.Lock()
        self._evt_closed = threading.Event()
        self._task_queue = collections.deque()
        self._pending_connection_futures = set()

        create_connection_future = self._register_pending_future()

        def on_open_error(conn, err):
            create_connection_future.set_exception(
                pika_exceptions.AMQPConnectionError(err)
            )

        self._impl = _impl_class(
            parameters=parameters,
            on_open_callback=create_connection_future.set_result,
            on_open_error_callback=on_open_error,
            on_close_callback=self._on_connection_close,
            stop_ioloop_on_close=False,
        )
        self._interrupt_pipein, self._interrupt_pipeout = os.pipe()
        self._impl.ioloop.add_handler(self._interrupt_pipein,
                                      self._impl.ioloop.read_interrupt,
                                      select_connection.READ)

        self._thread = threading.Thread(target=self._process_io)
        self._thread.daemon = True
        self._thread_id = None
        self._thread.start()

        create_connection_future.result()

    def _check_called_not_from_event_loop(self):
        if current_thread() == self._thread_id:
            raise RuntimeError("This call is not allowed from ioloop thread")

    def _execute_task(self, func, *args, **kwargs):
        if current_thread() == self._thread_id:
            return func(*args, **kwargs)

        future = futurist.Future()
        self._task_queue.append((func, args, kwargs, future))

        if self._evt_closed.is_set():
            self._notify_all_futures_connection_close()
        elif self._interrupt_pipeout is not None:
            os.write(self._interrupt_pipeout, b'X')

        return future.result()

    def _register_pending_future(self):
        future = futurist.Future()
        self._pending_connection_futures.add(future)

        def on_done_callback(fut):
            try:
                self._pending_connection_futures.remove(fut)
            except KeyError:
                pass

        future.add_done_callback(on_done_callback)

        if self._evt_closed.is_set():
            self._notify_all_futures_connection_close()
        return future

    def _notify_all_futures_connection_close(self):
        while self._task_queue:
            try:
                method_res_future = self._task_queue.pop()[3]
            except KeyError:
                break
            else:
                method_res_future.set_exception(
                    pika_exceptions.ConnectionClosed()
                )

        while self._pending_connection_futures:
            try:
                pending_connection_future = (
                    self._pending_connection_futures.pop()
                )
            except KeyError:
                break
            else:
                pending_connection_future.set_exception(
                    pika_exceptions.ConnectionClosed()
                )

    def _on_connection_close(self, conn, reply_code, reply_text):
        self._evt_closed.set()
        self._notify_all_futures_connection_close()
        if self._interrupt_pipeout:
            os.close(self._interrupt_pipeout)
            os.close(self._interrupt_pipein)

    def add_on_close_callback(self, callback):
        return self._execute_task(self._impl.add_on_close_callback, callback)

    def _do_process_io(self):
        while self._task_queue:
            func, args, kwargs, future = self._task_queue.pop()
            try:
                res = func(*args, **kwargs)
            except BaseException as e:
                LOG.exception(e)
                future.set_exception(e)
            else:
                future.set_result(res)

        self._impl.ioloop.poll()
        self._impl.ioloop.process_timeouts()

    def _process_io(self):
        self._thread_id = current_thread()
        while not self._evt_closed.is_set():
            try:
                self._do_process_io()
            except BaseException:
                LOG.exception("Error during processing connection's IO")

    def close(self, *args, **kwargs):
        self._check_called_not_from_event_loop()

        res = self._execute_task(self._impl.close, *args, **kwargs)

        self._evt_closed.wait()
        self._thread.join()
        return res

    def channel(self, channel_number=None):
        self._check_called_not_from_event_loop()

        channel_opened_future = self._register_pending_future()

        impl_channel = self._execute_task(
            self._impl.channel,
            on_open_callback=channel_opened_future.set_result,
            channel_number=channel_number
        )

        # Create our proxy channel
        channel = ThreadSafePikaChannel(impl_channel, self)

        # Link implementation channel with our proxy channel
        impl_channel._set_cookie(channel)

        channel_opened_future.result()
        return channel

    def add_timeout(self, timeout, callback):
        return self._execute_task(self._impl.add_timeout, timeout, callback)

    def remove_timeout(self, timeout_id):
        return self._execute_task(self._impl.remove_timeout, timeout_id)

    @property
    def is_closed(self):
        return self._impl.is_closed

    @property
    def is_closing(self):
        return self._impl.is_closing

    @property
    def is_open(self):
        return self._impl.is_open


class ThreadSafePikaChannel(object):  # pylint: disable=R0904,R0902

    def __init__(self, channel_impl, connection):
        self._impl = channel_impl
        self._connection = connection

        self._delivery_confirmation = False

        self._message_returned = False
        self._current_future = None

        self._evt_closed = threading.Event()

        self.add_on_close_callback(self._on_channel_close)

    def _execute_task(self, func, *args, **kwargs):
        return self._connection._execute_task(func, *args, **kwargs)

    def _on_channel_close(self, channel, reply_code, reply_text):
        self._evt_closed.set()

        if self._current_future:
            self._current_future.set_exception(
                pika_exceptions.ChannelClosed(reply_code, reply_text))

    def _on_message_confirmation(self, frame):
        self._current_future.set_result(frame)

    def add_on_close_callback(self, callback):
        self._execute_task(self._impl.add_on_close_callback, callback)

    def add_on_cancel_callback(self, callback):
        self._execute_task(self._impl.add_on_cancel_callback, callback)

    def __int__(self):
        return self.channel_number

    @property
    def channel_number(self):
        return self._impl.channel_number

    @property
    def is_closed(self):
        return self._impl.is_closed

    @property
    def is_closing(self):
        return self._impl.is_closing

    @property
    def is_open(self):
        return self._impl.is_open

    def close(self, reply_code=0, reply_text="Normal Shutdown"):
        self._impl.close(reply_code=reply_code, reply_text=reply_text)
        self._evt_closed.wait()

    def _check_called_not_from_event_loop(self):
        self._connection._check_called_not_from_event_loop()

    def flow(self, active):
        self._check_called_not_from_event_loop()

        self._current_future = futurist.Future()
        self._execute_task(
            self._impl.flow, callback=self._current_future.set_result,
            active=active
        )

        return self._current_future.result()

    def basic_consume(self,  # pylint: disable=R0913
                      consumer_callback,
                      queue,
                      no_ack=False,
                      exclusive=False,
                      consumer_tag=None,
                      arguments=None):

        self._check_called_not_from_event_loop()

        self._current_future = futurist.Future()
        self._execute_task(
            self._impl.add_callback, self._current_future.set_result,
            replies=[pika_spec.Basic.ConsumeOk], one_shot=True
        )

        self._impl.add_callback(self._current_future.set_result,
                                replies=[pika_spec.Basic.ConsumeOk],
                                one_shot=True)
        tag = self._execute_task(
            self._impl.basic_consume,
            consumer_callback=consumer_callback,
            queue=queue,
            no_ack=no_ack,
            exclusive=exclusive,
            consumer_tag=consumer_tag,
            arguments=arguments
        )

        self._current_future.result()
        return tag

    def basic_cancel(self, consumer_tag):
        self._check_called_not_from_event_loop()

        self._current_future = futurist.Future()
        self._execute_task(
            self._impl.basic_cancel,
            callback=self._current_future.set_result,
            consumer_tag=consumer_tag,
            nowait=False)
        self._current_future.result()

    def basic_ack(self, delivery_tag=0, multiple=False):
        return self._execute_task(
            self._impl.basic_ack, delivery_tag=delivery_tag, multiple=multiple)

    def basic_nack(self, delivery_tag=None, multiple=False, requeue=True):
        return self._execute_task(
            self._impl.basic_nack, delivery_tag=delivery_tag,
            multiple=multiple, requeue=requeue
        )

    def publish(self, exchange, routing_key, body,  # pylint: disable=R0913
                properties=None, mandatory=False, immediate=False):

        if self._delivery_confirmation:
            self._check_called_not_from_event_loop()

            # In publisher-acknowledgments mode
            self._message_returned = False
            self._current_future = futurist.Future()

            self._execute_task(self._impl.basic_publish,
                               exchange=exchange,
                               routing_key=routing_key,
                               body=body,
                               properties=properties,
                               mandatory=mandatory,
                               immediate=immediate)

            conf_method = self._current_future.result().method

            if isinstance(conf_method, pika_spec.Basic.Nack):
                raise pika_exceptions.NackError((None,))
            else:
                assert isinstance(conf_method, pika_spec.Basic.Ack), (
                    conf_method)

                if self._message_returned:
                    raise pika_exceptions.UnroutableError((None,))
        else:
            # In non-publisher-acknowledgments mode
            self._execute_task(self._impl.basic_publish,
                               exchange=exchange,
                               routing_key=routing_key,
                               body=body,
                               properties=properties,
                               mandatory=mandatory,
                               immediate=immediate)

    def basic_qos(self, prefetch_size=0, prefetch_count=0, all_channels=False):
        self._check_called_not_from_event_loop()

        self._current_future = futurist.Future()
        self._execute_task(self._impl.basic_qos,
                           callback=self._current_future.set_result,
                           prefetch_size=prefetch_size,
                           prefetch_count=prefetch_count,
                           all_channels=all_channels)
        self._current_future.result()

    def basic_recover(self, requeue=False):
        self._check_called_not_from_event_loop()

        self._current_future = futurist.Future()
        self._execute_task(
            self._impl.basic_recover,
            callback=lambda: self._current_future.set_result(None),
            requeue=requeue
        )
        self._current_future.result()

    def basic_reject(self, delivery_tag=None, requeue=True):
        self._execute_task(self._impl.basic_reject,
                           delivery_tag=delivery_tag,
                           requeue=requeue)

    def _on_message_returned(self, *args, **kwargs):
        self._message_returned = True

    def confirm_delivery(self):
        self._check_called_not_from_event_loop()

        self._current_future = futurist.Future()
        self._execute_task(self._impl.add_callback,
                           callback=self._current_future.set_result,
                           replies=[pika_spec.Confirm.SelectOk],
                           one_shot=True)
        self._execute_task(self._impl.confirm_delivery,
                           callback=self._on_message_confirmation,
                           nowait=False)
        self._current_future.result()

        self._delivery_confirmation = True
        self._execute_task(self._impl.add_on_return_callback,
                           self._on_message_returned)

    def exchange_declare(self, exchange=None,  # pylint: disable=R0913
                         exchange_type='direct', passive=False, durable=False,
                         auto_delete=False, internal=False,
                         arguments=None, **kwargs):
        self._check_called_not_from_event_loop()

        self._current_future = futurist.Future()
        self._execute_task(self._impl.exchange_declare,
                           callback=self._current_future.set_result,
                           exchange=exchange,
                           exchange_type=exchange_type,
                           passive=passive,
                           durable=durable,
                           auto_delete=auto_delete,
                           internal=internal,
                           nowait=False,
                           arguments=arguments,
                           type=kwargs["type"] if kwargs else None)

        return self._current_future.result()

    def exchange_delete(self, exchange=None, if_unused=False):
        self._check_called_not_from_event_loop()

        self._current_future = futurist.Future()
        self._execute_task(self._impl.exchange_delete,
                           callback=self._current_future.set_result,
                           exchange=exchange,
                           if_unused=if_unused,
                           nowait=False)

        return self._current_future.result()

    def exchange_bind(self, destination=None, source=None, routing_key='',
                      arguments=None):
        self._check_called_not_from_event_loop()

        self._current_future = futurist.Future()
        self._execute_task(self._impl.exchange_bind,
                           callback=self._current_future.set_result,
                           destination=destination,
                           source=source,
                           routing_key=routing_key,
                           nowait=False,
                           arguments=arguments)

        return self._current_future.result()

    def exchange_unbind(self, destination=None, source=None, routing_key='',
                        arguments=None):
        self._check_called_not_from_event_loop()

        self._current_future = futurist.Future()
        self._execute_task(self._impl.exchange_unbind,
                           callback=self._current_future.set_result,
                           destination=destination,
                           source=source,
                           routing_key=routing_key,
                           nowait=False,
                           arguments=arguments)

        return self._current_future.result()

    def queue_declare(self, queue='', passive=False, durable=False,
                      exclusive=False, auto_delete=False,
                      arguments=None):
        self._check_called_not_from_event_loop()

        self._current_future = futurist.Future()
        self._execute_task(self._impl.queue_declare,
                           callback=self._current_future.set_result,
                           queue=queue,
                           passive=passive,
                           durable=durable,
                           exclusive=exclusive,
                           auto_delete=auto_delete,
                           nowait=False,
                           arguments=arguments)

        return self._current_future.result()

    def queue_delete(self, queue='', if_unused=False, if_empty=False):
        self._check_called_not_from_event_loop()

        self._current_future = futurist.Future()
        self._execute_task(self._impl.queue_delete,
                           callback=self._current_future.set_result,
                           queue=queue,
                           if_unused=if_unused,
                           if_empty=if_empty,
                           nowait=False)

        return self._current_future.result()

    def queue_purge(self, queue=''):
        self._check_called_not_from_event_loop()

        self._current_future = futurist.Future()
        self._execute_task(self._impl.queue_purge,
                           callback=self._current_future.set_result,
                           queue=queue,
                           nowait=False)
        return self._current_future.result()

    def queue_bind(self, queue, exchange, routing_key=None,
                   arguments=None):
        self._check_called_not_from_event_loop()

        self._current_future = futurist.Future()
        self._execute_task(self._impl.queue_bind,
                           callback=self._current_future.set_result,
                           queue=queue,
                           exchange=exchange,
                           routing_key=routing_key,
                           nowait=False,
                           arguments=arguments)
        return self._current_future.result()

    def queue_unbind(self, queue='', exchange=None, routing_key=None,
                     arguments=None):
        self._check_called_not_from_event_loop()

        self._current_future = futurist.Future()
        self._execute_task(self._impl.queue_unbind,
                           callback=self._current_future.set_result,
                           queue=queue,
                           exchange=exchange,
                           routing_key=routing_key,
                           arguments=arguments)
        return self._current_future.result()



# Copyright 2017 OpenStack Foundation.
# All Rights Reserved.
# Copyright 2017 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
__all__ = [
    'get_rpc_transport'
]

from oslo_messaging import transport as msg_transport


def get_rpc_transport(conf, url=None,
                      allowed_remote_exmods=None):
    """A factory method for Transport objects for RPCs.

    This method should be used to ensure the correct messaging functionality
    for RPCs. RPCs and Notifications may use separate messaging systems
    that utilize different drivers, different access permissions,
    message delivery, etc.

    Presently, this function works exactly the same as get_transport. It's
    use is recommended as disambiguates the intended use for the transport
    and may in the future extend functionality related to the separation of
    messaging backends.

    :param conf: the user configuration
    :type conf: cfg.ConfigOpts
    :param url: a transport URL, see :py:class:`transport.TransportURL`
    :type url: str or TransportURL
    :param allowed_remote_exmods: a list of modules which a client using this
                                  transport will deserialize remote exceptions
                                  from
    :type allowed_remote_exmods: list
    """
    return msg_transport._get_transport(conf, url,
                                        allowed_remote_exmods)



#    Copyright 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import logging.handlers
import multiprocessing
import os
import sys
import threading
import time
import uuid

from oslo_config import cfg

import oslo_messaging
from oslo_messaging._drivers.zmq_driver import zmq_async
from oslo_messaging.tests.functional import utils

LOG = logging.getLogger(__name__)

zmq = zmq_async.import_zmq()


class QueueHandler(logging.Handler):
    """This is a logging handler which sends events to a multiprocessing queue.

    The plan is to add it to Python 3.2, but this can be copy pasted into
    user code for use with earlier Python versions.
    """

    def __init__(self, queue):
        """Initialise an instance, using the passed queue."""
        logging.Handler.__init__(self)
        self.queue = queue

    def emit(self, record):
        """Emit a record.

        Writes the LogRecord to the queue.
        """
        try:
            ei = record.exc_info
            if ei:
                # just to get traceback text into record.exc_text
                dummy = self.format(record)  # noqa
                record.exc_info = None  # not needed any more
            self.queue.put_nowait(record)
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception:
            self.handleError(record)


def listener_configurer(conf):
    root = logging.getLogger()
    h = logging.StreamHandler(sys.stdout)
    f = logging.Formatter('%(asctime)s %(processName)-10s %(name)s '
                          '%(levelname)-8s %(message)s')
    h.setFormatter(f)
    root.addHandler(h)
    log_path = conf.oslo_messaging_zmq.rpc_zmq_ipc_dir + \
        "/" + "zmq_multiproc.log"
    file_handler = logging.StreamHandler(open(log_path, 'w'))
    file_handler.setFormatter(f)
    root.addHandler(file_handler)


def server_configurer(queue):
    h = QueueHandler(queue)
    root = logging.getLogger()
    root.addHandler(h)
    root.setLevel(logging.DEBUG)


def listener_thread(queue, configurer, conf):
    configurer(conf)
    while True:
        time.sleep(0.3)
        try:
            record = queue.get()
            if record is None:
                break
            logger = logging.getLogger(record.name)
            logger.handle(record)
        except (KeyboardInterrupt, SystemExit):
            raise


class Client(oslo_messaging.RPCClient):

    def __init__(self, transport, topic):
        super(Client, self).__init__(
            transport=transport, target=oslo_messaging.Target(topic=topic))
        self.replies = []

    def call_a(self):
        LOG.warning("call_a - client side")
        rep = self.call({}, 'call_a')
        LOG.warning("after call_a - client side")
        self.replies.append(rep)
        return rep


class ReplyServerEndpoint(object):

    def call_a(self, *args, **kwargs):
        LOG.warning("call_a - Server endpoint reached!")
        return "OK"


class Server(object):

    def __init__(self, conf, log_queue, transport_url, name, topic=None):
        self.conf = conf
        self.log_queue = log_queue
        self.transport_url = transport_url
        self.name = name
        self.topic = topic or str(uuid.uuid4())
        self.ready = multiprocessing.Value('b', False)
        self._stop = multiprocessing.Event()

    def start(self):
        self.process = multiprocessing.Process(target=self._run_server,
                                               name=self.name,
                                               args=(self.conf,
                                                     self.transport_url,
                                                     self.log_queue,
                                                     self.ready))
        self.process.start()
        LOG.debug("Server process started: pid: %d", self.process.pid)

    def _run_server(self, conf, url, log_queue, ready):
        server_configurer(log_queue)
        LOG.debug("Starting RPC server")

        transport = oslo_messaging.get_transport(conf, url=url)
        target = oslo_messaging.Target(topic=self.topic, server=self.name)
        self.rpc_server = oslo_messaging.get_rpc_server(
            transport=transport, target=target,
            endpoints=[ReplyServerEndpoint()],
            executor='eventlet')
        self.rpc_server.start()
        ready.value = True
        LOG.debug("RPC server being started")
        while not self._stop.is_set():
            LOG.debug("Waiting for the stop signal ...")
            time.sleep(1)
        self.rpc_server.stop()
        self.rpc_server.wait()
        LOG.debug("Leaving process T:%s Pid:%d", str(target), os.getpid())

    def cleanup(self):
        LOG.debug("Stopping server")
        self.shutdown()

    def shutdown(self):
        self._stop.set()

    def restart(self, time_for_restart=1):
        pass

    def hang(self):
        pass

    def crash(self):
        pass

    def ping(self):
        pass


class MultiprocTestCase(utils.SkipIfNoTransportURL):

    def setUp(self):
        super(MultiprocTestCase, self).setUp(conf=cfg.ConfigOpts())

        if not self.url.startswith("zmq"):
            self.skipTest("ZeroMQ specific skipped...")

        self.transport = oslo_messaging.get_transport(self.conf, url=self.url)

        LOG.debug("Start log queue")

        self.log_queue = multiprocessing.Queue()
        self.log_listener = threading.Thread(target=listener_thread,
                                             args=(self.log_queue,
                                                   listener_configurer,
                                                   self.conf))
        self.log_listener.start()
        self.spawned = []

        self.conf.prog = "test_prog"
        self.conf.project = "test_project"

    def tearDown(self):
        for process in self.spawned:
            process.cleanup()
        super(MultiprocTestCase, self).tearDown()

    def get_client(self, topic):
        return Client(self.transport, topic)

    def spawn_server(self, wait_for_server=False, topic=None):
        name = "server_%d_%s" % (len(self.spawned), str(uuid.uuid4())[:8])
        server = Server(self.conf, self.log_queue, self.url, name, topic)
        LOG.debug("[SPAWN] %s (starting)...", server.name)
        server.start()
        if wait_for_server:
            while not server.ready.value:
                LOG.debug("[SPAWN] %s (waiting for server ready)...",
                          server.name)
                time.sleep(1)
        LOG.debug("[SPAWN] Server %s:%d started.",
                  server.name, server.process.pid)
        self.spawned.append(server)
        return server

    def spawn_servers(self, number, wait_for_server=False, common_topic=True):
        topic = str(uuid.uuid4()) if common_topic else None
        for _ in range(number):
            self.spawn_server(wait_for_server, topic)



#    Copyright 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import itertools
import logging
import threading
import time

from oslo_messaging._drivers.zmq_driver.matchmaker import zmq_matchmaker_base
from oslo_messaging._drivers.zmq_driver import zmq_address
from oslo_messaging._drivers.zmq_driver import zmq_async
from oslo_messaging._drivers.zmq_driver import zmq_names
from oslo_messaging._drivers.zmq_driver import zmq_updater
from oslo_messaging._i18n import _LW

LOG = logging.getLogger(__name__)

zmq = zmq_async.import_zmq()


class RoutingTableAdaptor(object):

    def __init__(self, conf, matchmaker, listener_type):
        self.conf = conf
        self.matchmaker = matchmaker
        self.listener_type = listener_type
        self.routing_table = RoutingTable(conf)
        self.routing_table_updater = RoutingTableUpdater(
            conf, matchmaker, self.routing_table)
        self.round_robin_targets = {}
        self._lock = threading.Lock()

    def get_round_robin_host(self, target):
        target_key = self._fetch_round_robin_hosts_from_matchmaker(target)
        rr_gen = self.round_robin_targets[target_key]
        host = next(rr_gen)
        LOG.debug("Host resolved for the current connection is %s" % host)
        return host

    def get_all_round_robin_hosts(self, target):
        target_key = self._fetch_round_robin_hosts_from_matchmaker(target)
        return self.routing_table.get_hosts_fanout(target_key)

    def _fetch_round_robin_hosts_from_matchmaker(self, target):
        target_key = zmq_address.target_to_key(
            target, zmq_names.socket_type_str(self.listener_type))

        LOG.debug("Processing target %s for round-robin." % target_key)

        if target_key not in self.round_robin_targets:
            with self._lock:
                if target_key not in self.round_robin_targets:
                    LOG.debug("Target %s is not in cache. Check matchmaker "
                              "server." % target_key)
                    hosts = self.matchmaker.get_hosts_retry(
                        target, zmq_names.socket_type_str(self.listener_type))
                    LOG.debug("Received hosts %s" % hosts)
                    self.routing_table.update_hosts(target_key, hosts)
                    self.round_robin_targets[target_key] = \
                        self.routing_table.get_hosts_round_robin(target_key)
        return target_key

    def get_fanout_hosts(self, target):
        target_key = zmq_address.target_to_key(
            target, zmq_names.socket_type_str(self.listener_type))

        LOG.debug("Processing target %s for fanout." % target_key)

        if not self.routing_table.contains(target_key):
            self._fetch_fanout_hosts_from_matchmaker(target, target_key)

        return self.routing_table.get_hosts_fanout(target_key)

    def _fetch_fanout_hosts_from_matchmaker(self, target, target_key):
        with self._lock:
            if not self.routing_table.contains(target_key):
                LOG.debug("Target %s is not in cache. Check matchmaker server."
                          % target_key)
                hosts = self.matchmaker.get_hosts_fanout(
                    target, zmq_names.socket_type_str(self.listener_type))
                LOG.debug("Received hosts %s" % hosts)
                self.routing_table.update_hosts(target_key, hosts)

    def cleanup(self):
        self.routing_table_updater.cleanup()


class RoutingTable(object):

    def __init__(self, conf):
        self.conf = conf
        self.targets = {}
        self._lock = threading.Lock()

    def register(self, target_key, host):
        with self._lock:
            if target_key in self.targets:
                hosts, tm = self.targets[target_key]
                if host not in hosts:
                    hosts.add(host)
                    self.targets[target_key] = (hosts, self._create_tm())
            else:
                self.targets[target_key] = ({host}, self._create_tm())

    def get_targets(self):
        with self._lock:
            return list(self.targets.keys())

    def unregister(self, target_key, host):
        with self._lock:
            hosts, tm = self.targets.get(target_key)
            if hosts and host in hosts:
                hosts.discard(host)
                self.targets[target_key] = (hosts, self._create_tm())

    def update_hosts(self, target_key, hosts_updated):
        with self._lock:
            if target_key in self.targets and not hosts_updated:
                self.targets.pop(target_key)
                return
            hosts_current, _ = self.targets.get(target_key, (set(), None))
            hosts_updated = set(hosts_updated)
            has_differences = hosts_updated ^ hosts_current
            if has_differences:
                self.targets[target_key] = (hosts_updated, self._create_tm())

    def get_hosts_round_robin(self, target_key):
        while self.contains(target_key):
            for host in self._get_hosts_rr(target_key):
                yield host

    def get_hosts_fanout(self, target_key):
        hosts, _ = self._get_hosts(target_key)
        return hosts

    def contains(self, target_key):
        with self._lock:
            return target_key in self.targets

    def _get_hosts(self, target_key):
        with self._lock:
            hosts, tm = self.targets.get(target_key, ([], None))
            hosts = list(hosts)
            return hosts, tm

    def _get_tm(self, target_key):
        with self._lock:
            _, tm = self.targets.get(target_key)
            return tm

    def _is_target_changed(self, target_key, tm_orig):
        return self._get_tm(target_key) != tm_orig

    @staticmethod
    def _create_tm():
        return time.time()

    def _get_hosts_rr(self, target_key):
        hosts, tm_original = self._get_hosts(target_key)
        for host in itertools.cycle(hosts):
            if self._is_target_changed(target_key, tm_original):
                raise StopIteration()
            yield host


class RoutingTableUpdater(zmq_updater.UpdaterBase):

    def __init__(self, conf, matchmaker, routing_table):
        self.routing_table = routing_table
        super(RoutingTableUpdater, self).__init__(
            conf, matchmaker, self._update_routing_table,
            conf.oslo_messaging_zmq.zmq_target_update)

    def _update_routing_table(self):
        target_keys = self.routing_table.get_targets()

        try:
            for target_key in target_keys:
                hosts = self.matchmaker.get_hosts_by_key(target_key)
                self.routing_table.update_hosts(target_key, hosts)
            LOG.debug("Updating routing table from the matchmaker. "
                      "%d target(s) updated %s." % (len(target_keys),
                                                    target_keys))
        except zmq_matchmaker_base.MatchmakerUnavailable:
            LOG.warning(_LW("Not updated. Matchmaker was not available."))



#    Copyright 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import uuid

import six

from oslo_messaging._drivers import common as rpc_common
from oslo_messaging._drivers.zmq_driver.client import zmq_response
from oslo_messaging._drivers.zmq_driver.client import zmq_senders
from oslo_messaging._drivers.zmq_driver.client import zmq_sockets_manager
from oslo_messaging._drivers.zmq_driver.server.consumers \
    import zmq_consumer_base
from oslo_messaging._drivers.zmq_driver.server import zmq_incoming_message
from oslo_messaging._drivers.zmq_driver.server import zmq_ttl_cache
from oslo_messaging._drivers.zmq_driver import zmq_address
from oslo_messaging._drivers.zmq_driver import zmq_async
from oslo_messaging._drivers.zmq_driver import zmq_names
from oslo_messaging._drivers.zmq_driver import zmq_updater
from oslo_messaging._drivers.zmq_driver import zmq_version
from oslo_messaging._i18n import _LE, _LI, _LW

LOG = logging.getLogger(__name__)

zmq = zmq_async.import_zmq()


class DealerConsumer(zmq_consumer_base.SingleSocketConsumer):

    def __init__(self, conf, poller, server):
        self.reply_sender = zmq_senders.ReplySenderProxy(conf)
        self.sockets_manager = zmq_sockets_manager.SocketsManager(
            conf, server.matchmaker, zmq.DEALER)
        self.host = None
        super(DealerConsumer, self).__init__(conf, poller, server, zmq.DEALER)
        self._receive_request_versions = \
            zmq_version.get_method_versions(self, 'receive_request')
        self.connection_updater = ConsumerConnectionUpdater(
            conf, self.matchmaker, self.socket)
        LOG.info(_LI("[%s] Run DEALER consumer"), self.host)

    def _generate_identity(self):
        return six.b(self.conf.oslo_messaging_zmq.rpc_zmq_host + "/" +
                     zmq_address.target_to_key(self.target) + "/" +
                     str(uuid.uuid4()))

    def subscribe_socket(self, socket_type):
        try:
            socket = self.sockets_manager.get_socket_to_routers(
                self._generate_identity())
            self.host = socket.handle.identity
            self.poller.register(socket, self.receive_request)
            return socket
        except zmq.ZMQError as e:
            LOG.error(_LE("Failed connecting to ROUTER socket %(e)s") % e)
            raise rpc_common.RPCException(str(e))

    def _reply(self, rpc_message, reply, failure):
        if failure is not None:
            failure = rpc_common.serialize_remote_exception(failure)
        reply = zmq_response.Reply(message_id=rpc_message.message_id,
                                   reply_id=rpc_message.reply_id,
                                   message_version=rpc_message.message_version,
                                   reply_body=reply,
                                   failure=failure)
        self.reply_sender.send(rpc_message.socket, reply)
        return reply

    def _create_message(self, context, message, message_version, reply_id,
                        message_id, socket, message_type):
        if message_type == zmq_names.CALL_TYPE:
            message = zmq_incoming_message.ZmqIncomingMessage(
                context, message, message_version=message_version,
                reply_id=reply_id, message_id=message_id,
                socket=socket, reply_method=self._reply
            )
        else:
            message = zmq_incoming_message.ZmqIncomingMessage(context, message)

        LOG.debug("[%(host)s] Received %(msg_type)s message %(msg_id)s "
                  "(v%(msg_version)s)",
                  {"host": self.host,
                   "msg_type": zmq_names.message_type_str(message_type),
                   "msg_id": message_id,
                   "msg_version": message_version})
        return message

    def _get_receive_request_version(self, version):
        receive_request_version = self._receive_request_versions.get(version)
        if receive_request_version is None:
            raise zmq_version.UnsupportedMessageVersionError(version)
        return receive_request_version

    def receive_request(self, socket):
        try:
            empty = socket.recv()
            assert empty == b'', "Empty delimiter expected!"
            message_version = socket.recv_string()
            assert message_version != b'', "Valid message version expected!"

            receive_request_version = \
                self._get_receive_request_version(message_version)
            return receive_request_version(socket)
        except (zmq.ZMQError, AssertionError, ValueError,
                zmq_version.UnsupportedMessageVersionError) as e:
            LOG.error(_LE("Receiving message failure: %s"), str(e))
            # NOTE(gdavoian): drop the left parts of a broken message
            if socket.getsockopt(zmq.RCVMORE):
                socket.recv_multipart()

    def _receive_request_v_1_0(self, socket):
        reply_id = socket.recv()
        assert reply_id != b'', "Valid reply id expected!"
        message_type = int(socket.recv())
        assert message_type in zmq_names.REQUEST_TYPES, "Request expected!"
        message_id = socket.recv_string()
        assert message_id != '', "Valid message id expected!"
        context, message = socket.recv_loaded()

        return self._create_message(context, message, '1.0', reply_id,
                                    message_id, socket, message_type)

    def cleanup(self):
        LOG.info(_LI("[%s] Destroy DEALER consumer"), self.host)
        self.connection_updater.cleanup()
        super(DealerConsumer, self).cleanup()


class DealerConsumerWithAcks(DealerConsumer):

    def __init__(self, conf, poller, server):
        super(DealerConsumerWithAcks, self).__init__(conf, poller, server)
        self.ack_sender = zmq_senders.AckSenderProxy(conf)
        self.messages_cache = zmq_ttl_cache.TTLCache(
            ttl=conf.oslo_messaging_zmq.rpc_message_ttl
        )

    def _acknowledge(self, message_version, reply_id, message_id, socket):
        ack = zmq_response.Ack(message_id=message_id,
                               reply_id=reply_id,
                               message_version=message_version)
        self.ack_sender.send(socket, ack)

    def _reply(self, rpc_message, reply, failure):
        reply = super(DealerConsumerWithAcks, self)._reply(rpc_message,
                                                           reply, failure)
        self.messages_cache.add(rpc_message.message_id, reply)
        return reply

    def _reply_from_cache(self, message_id, socket):
        reply = self.messages_cache.get(message_id)
        if reply is not None:
            self.reply_sender.send(socket, reply)

    def _create_message(self, context, message, message_version, reply_id,
                        message_id, socket, message_type):
        # drop a duplicate message
        if message_id in self.messages_cache:
            LOG.warning(
                _LW("[%(host)s] Dropping duplicate %(msg_type)s "
                    "message %(msg_id)s"),
                {"host": self.host,
                 "msg_type": zmq_names.message_type_str(message_type),
                 "msg_id": message_id}
            )
            # NOTE(gdavoian): send yet another ack for the direct
            # message, since the old one might be lost;
            # for the CALL message also try to resend its reply
            # (of course, if it was already obtained and cached).
            if message_type in zmq_names.DIRECT_TYPES:
                self._acknowledge(message_version, reply_id, message_id,
                                  socket)
            if message_type == zmq_names.CALL_TYPE:
                self._reply_from_cache(message_id, socket)
            return None

        self.messages_cache.add(message_id)

        # NOTE(gdavoian): send an immediate ack, since it may
        # be too late to wait until the message will be
        # dispatched and processed by a RPC server
        if message_type in zmq_names.DIRECT_TYPES:
            self._acknowledge(message_version, reply_id, message_id, socket)

        return super(DealerConsumerWithAcks, self)._create_message(
            context, message, message_version, reply_id,
            message_id, socket, message_type
        )

    def cleanup(self):
        self.messages_cache.cleanup()
        super(DealerConsumerWithAcks, self).cleanup()


class ConsumerConnectionUpdater(zmq_updater.ConnectionUpdater):

    def _update_connection(self):
        routers = self.matchmaker.get_routers()
        for router_address in routers:
            self.socket.connect_to_host(router_address)



#    Copyright 2015-2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

import tenacity

from oslo_messaging._drivers.zmq_driver.client.publishers.dealer \
    import zmq_dealer_publisher_base
from oslo_messaging._drivers.zmq_driver.client import zmq_receivers
from oslo_messaging._drivers.zmq_driver.client import zmq_routing_table
from oslo_messaging._drivers.zmq_driver.client import zmq_senders
from oslo_messaging._drivers.zmq_driver.client import zmq_sockets_manager
from oslo_messaging._drivers.zmq_driver import zmq_address
from oslo_messaging._drivers.zmq_driver import zmq_async
from oslo_messaging._drivers.zmq_driver import zmq_names


LOG = logging.getLogger(__name__)

zmq = zmq_async.import_zmq()


class DealerPublisherDirect(zmq_dealer_publisher_base.DealerPublisherBase):
    """DEALER-publisher using direct dynamic connections.

    Publishing directly to remote services assumes the following:
        -   All direct connections are dynamic - so they live per message,
            thus each message send executes the following:
                *   Open a new socket
                *   Connect to some host got from the RoutingTable
                *   Send message(s)
                *   Close connection, destroy socket
        -   RoutingTable/RoutingTableUpdater implements local cache of
            matchmaker (e.g. Redis) for target resolution to the list of
            available hosts. Cache updates in a background thread.
        -   Caching of connections is not appropriate for directly connected
            OS services, because finally it results in a full-mesh of
            connections between services.
        -   Yes we lose on performance opening and closing connections
            for each message, but that is done intentionally to implement
            the dynamic connections concept. The key thought here is to
            have minimum number of connected services at the moment.
        -   Using the local RoutingTable cache is done to optimise access
            to the matchmaker so we don't call the matchmaker per each message
    """

    def __init__(self, conf, matchmaker):
        sender = zmq_senders.RequestSenderDirect(conf, async=True)
        receiver = zmq_receivers.ReceiverDirect(conf)
        super(DealerPublisherDirect, self).__init__(conf, matchmaker,
                                                    sender, receiver)

        self.routing_table = zmq_routing_table.RoutingTableAdaptor(
            conf, matchmaker, zmq.ROUTER)

    def _get_round_robin_host_connection(self, target, socket):
        host = self.routing_table.get_round_robin_host(target)
        socket.connect_to_host(host)
        failover_hosts = self.routing_table.get_all_round_robin_hosts(target)
        upper_bound = self.conf.oslo_messaging_zmq.zmq_failover_connections
        for host in failover_hosts[:upper_bound]:
            socket.connect_to_host(host)

    def _get_fanout_connection(self, target, socket):
        for host in self.routing_table.get_fanout_hosts(target):
            socket.connect_to_host(host)

    def acquire_connection(self, request):
        if request.msg_type in zmq_names.DIRECT_TYPES:
            socket = self.sockets_manager.get_socket()
            self._get_round_robin_host_connection(request.target, socket)
            return socket
        elif request.msg_type in zmq_names.MULTISEND_TYPES:
            socket = self.sockets_manager.get_socket(immediate=False)
            self._get_fanout_connection(request.target, socket)
            return socket

    def _finally_unregister(self, socket, request):
        super(DealerPublisherDirect, self)._finally_unregister(socket, request)
        self.receiver.unregister_socket(socket)

    def send_request(self, socket, request):
        if hasattr(request, 'timeout'):
            _stop = tenacity.stop_after_delay(request.timeout)
        elif request.retry is not None and request.retry > 0:
            # no rpc_response_timeout option if notification
            _stop = tenacity.stop_after_attempt(request.retry)
        else:
            # well, now what?
            _stop = tenacity.stop_after_delay(60)

        @tenacity.retry(retry=tenacity.retry_if_exception_type(zmq.Again),
                        stop=_stop)
        def send_retrying():
            if request.msg_type in zmq_names.MULTISEND_TYPES:
                for _ in range(socket.connections_count()):
                    self.sender.send(socket, request)
            else:
                self.sender.send(socket, request)
        return send_retrying()

    def cleanup(self):
        self.routing_table.cleanup()
        super(DealerPublisherDirect, self).cleanup()


class DealerPublisherDirectStatic(DealerPublisherDirect):
    """DEALER-publisher using direct static connections.

    For some reason direct static connections may be also useful.
    Assume a case when some agents are not connected with control services
    over RPC (Ironic or Cinder+Ceph), and RPC is used only between controllers.
    In this case number of RPC connections doesn't matter (very small) so we
    can use static connections without fear and have all performance benefits
    from it.
    """

    def __init__(self, conf, matchmaker):
        super(DealerPublisherDirectStatic, self).__init__(conf, matchmaker)
        self.fanout_sockets = zmq_sockets_manager.SocketsManager(
            conf, matchmaker, zmq.DEALER)

    def acquire_connection(self, request):
        target_key = zmq_address.target_to_key(
            request.target, zmq_names.socket_type_str(zmq.ROUTER))
        if request.msg_type in zmq_names.MULTISEND_TYPES:
            hosts = self.routing_table.get_fanout_hosts(request.target)
            return self.fanout_sockets.get_cached_socket(target_key, hosts,
                                                         immediate=False)
        else:
            hosts = self.routing_table.get_all_round_robin_hosts(
                request.target)
            return self.sockets_manager.get_cached_socket(target_key, hosts)

    def _finally_unregister(self, socket, request):
        self.receiver.untrack_request(request)

    def cleanup(self):
        self.fanout_sockets.cleanup()
        super(DealerPublisherDirectStatic, self).cleanup()



# Copyright 2014 Canonical, Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import inspect
from stevedore import driver
import testscenarios

import oslo_messaging
from oslo_messaging.tests import utils as test_utils
from oslo_utils import importutils

redis = importutils.try_import('redis')


def redis_available():
    '''Helper to see if local redis server is running'''
    if not redis:
        return False
    try:
        redis.StrictRedis(socket_timeout=1).ping()
        return True
    except redis.exceptions.ConnectionError:
        return False


load_tests = testscenarios.load_tests_apply_scenarios


class TestImplMatchmaker(test_utils.BaseTestCase):

    scenarios = [
        ("dummy", {"rpc_zmq_matchmaker": "dummy"}),
        ("redis", {"rpc_zmq_matchmaker": "redis"}),
    ]

    def setUp(self):
        super(TestImplMatchmaker, self).setUp()

        if self.rpc_zmq_matchmaker == "redis":
            if not redis_available():
                self.skipTest("redis unavailable")

        self.test_matcher = driver.DriverManager(
            'oslo.messaging.zmq.matchmaker',
            self.rpc_zmq_matchmaker,
        ).driver(self.conf)

        if self.rpc_zmq_matchmaker == "redis":
            for redis_instance in self.test_matcher._redis_instances:
                self.addCleanup(redis_instance.flushdb)

        self.target = oslo_messaging.Target(topic="test_topic")
        self.host1 = b"test_host1"
        self.host2 = b"test_host2"

    def test_register(self):
        self.test_matcher.register(
            self.target,
            self.host1,
            "test",
            expire=self.conf.oslo_messaging_zmq.zmq_target_expire)

        self.assertEqual([self.host1],
                         self.test_matcher.get_hosts(self.target, "test"))

    def test_register_two_hosts(self):
        self.test_matcher.register(
            self.target,
            self.host1,
            "test",
            expire=self.conf.oslo_messaging_zmq.zmq_target_expire)
        self.test_matcher.register(
            self.target,
            self.host2,
            "test",
            expire=self.conf.oslo_messaging_zmq.zmq_target_expire)

        self.assertItemsEqual(self.test_matcher.get_hosts(self.target, "test"),
                              [self.host1, self.host2])

    def test_register_unregister(self):
        self.test_matcher.register(
            self.target,
            self.host1,
            "test",
            expire=self.conf.oslo_messaging_zmq.zmq_target_expire)
        self.test_matcher.register(
            self.target,
            self.host2,
            "test",
            expire=self.conf.oslo_messaging_zmq.zmq_target_expire)

        self.test_matcher.unregister(self.target, self.host2, "test")

        self.assertItemsEqual(self.test_matcher.get_hosts(self.target, "test"),
                              [self.host1])

    def test_register_two_same_hosts(self):
        self.test_matcher.register(
            self.target,
            self.host1,
            "test",
            expire=self.conf.oslo_messaging_zmq.zmq_target_expire)
        self.test_matcher.register(
            self.target,
            self.host1,
            "test",
            expire=self.conf.oslo_messaging_zmq.zmq_target_expire)

        self.assertEqual([self.host1],
                         self.test_matcher.get_hosts(self.target, "test"))

    def test_get_hosts_wrong_topic(self):
        target = oslo_messaging.Target(topic="no_such_topic")
        self.assertEqual([], self.test_matcher.get_hosts(target, "test"))

    def test_handle_redis_package_error(self):
        if self.rpc_zmq_matchmaker == "redis":
            # move 'redis' variable to prevent this case affect others
            module = inspect.getmodule(self.test_matcher)
            redis_package = module.redis

            # 'redis' variable is set to None, when package importing is failed
            module.redis = None
            self.assertRaises(ImportError, self.test_matcher.__init__,
                              self.conf)

            # retrieve 'redis' variable which is set originally
            module.redis = redis_package



#    Copyright 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import logging

import six

from oslo_messaging._drivers import common as rpc_common
from oslo_messaging._drivers.zmq_driver.matchmaker import zmq_matchmaker_base
from oslo_messaging._drivers.zmq_driver import zmq_address
from oslo_messaging._drivers.zmq_driver import zmq_async
from oslo_messaging._drivers.zmq_driver import zmq_names
from oslo_messaging._drivers.zmq_driver import zmq_socket
from oslo_messaging._drivers.zmq_driver import zmq_updater
from oslo_messaging._i18n import _LE, _LI, _LW

LOG = logging.getLogger(__name__)

zmq = zmq_async.import_zmq()


@six.add_metaclass(abc.ABCMeta)
class ConsumerBase(object):

    def __init__(self, conf, poller, server):
        self.conf = conf
        self.poller = poller
        self.server = server
        self.sockets = []
        self.context = zmq.Context()

    def stop(self):
        """Stop consumer polling/updating."""

    @abc.abstractmethod
    def receive_request(self, socket):
        """Receive a request via a socket."""

    def cleanup(self):
        for socket in self.sockets:
            if not socket.handle.closed:
                socket.close()
        self.sockets = []


class SingleSocketConsumer(ConsumerBase):

    def __init__(self, conf, poller, server, socket_type):
        super(SingleSocketConsumer, self).__init__(conf, poller, server)
        self.matchmaker = server.matchmaker
        self.target = server.target
        self.socket_type = socket_type
        self.host = None
        self.socket = self.subscribe_socket(socket_type)
        self.target_updater = TargetUpdater(
            conf, self.matchmaker, self.target, self.host, socket_type)

    def stop(self):
        self.target_updater.stop()

    def subscribe_socket(self, socket_type):
        try:
            socket = zmq_socket.ZmqRandomPortSocket(
                self.conf, self.context, socket_type)
            self.sockets.append(socket)
            LOG.debug("Run %(stype)s consumer on %(addr)s:%(port)d",
                      {"stype": zmq_names.socket_type_str(socket_type),
                       "addr": socket.bind_address,
                       "port": socket.port})
            self.host = zmq_address.combine_address(
                self.conf.oslo_messaging_zmq.rpc_zmq_host, socket.port)
            self.poller.register(socket, self.receive_request)
            return socket
        except zmq.ZMQError as e:
            errmsg = _LE("Failed binding to port %(port)d: %(e)s")\
                % (self.port, e)
            LOG.error(_LE("Failed binding to port %(port)d: %(e)s"),
                      (self.port, e))
            raise rpc_common.RPCException(errmsg)

    @property
    def address(self):
        return self.socket.bind_address

    @property
    def port(self):
        return self.socket.port

    def cleanup(self):
        self.target_updater.cleanup()
        super(SingleSocketConsumer, self).cleanup()


class TargetUpdater(zmq_updater.UpdaterBase):
    """This entity performs periodic async updates
    to the matchmaker.
    """

    def __init__(self, conf, matchmaker, target, host, socket_type):
        self.target = target
        self.host = host
        self.socket_type = socket_type
        self.conf = conf
        self.matchmaker = matchmaker
        self._sleep_for = conf.oslo_messaging_zmq.zmq_target_update

        # NOTE(ozamiatin): Update target immediately not waiting
        # for background executor to initialize.
        self._update_target()

        super(TargetUpdater, self).__init__(
            conf, matchmaker, self._update_target,
            conf.oslo_messaging_zmq.zmq_target_update)

    def _update_target(self):
        try:
            self.matchmaker.register(
                self.target, self.host,
                zmq_names.socket_type_str(self.socket_type),
                expire=self.conf.oslo_messaging_zmq.zmq_target_expire)

            if self._sleep_for != \
                    self.conf.oslo_messaging_zmq.zmq_target_update:
                self._sleep_for = \
                    self.conf.oslo_messaging_zmq.zmq_target_update
                LOG.info(_LI("Falling back to the normal update %d sec")
                         % self._sleep_for)

        except zmq_matchmaker_base.MatchmakerUnavailable:
            # Update target frequently until first successful update
            # After matchmaker is back update normally as of config
            self._sleep_for = 10
            LOG.warning(_LW("Failed connecting to the Matchmaker, "
                            "update each %d sec") % self._sleep_for)

    def stop(self):
        super(TargetUpdater, self).stop()
        self.matchmaker.unregister(
            self.target, self.host,
            zmq_names.socket_type_str(self.socket_type))



#    Copyright 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import sys

from oslo_messaging.tests.functional.zmq import multiproc_utils


class StartupOrderTestCase(multiproc_utils.MultiprocTestCase):

    def setUp(self):
        super(StartupOrderTestCase, self).setUp()

        self.conf.prog = "test_prog"
        self.conf.project = "test_project"

        self.config(rpc_response_timeout=10)

        log_path = os.path.join(self.conf.oslo_messaging_zmq.rpc_zmq_ipc_dir,
                                str(os.getpid()) + ".log")
        sys.stdout = open(log_path, "wb", buffering=0)

    def test_call_client_wait_for_server(self):
        server = self.spawn_server(wait_for_server=True)
        client = self.get_client(server.topic)
        for _ in range(3):
            reply = client.call_a()
            self.assertIsNotNone(reply)
        self.assertEqual(3, len(client.replies))

    def test_call_client_dont_wait_for_server(self):
        server = self.spawn_server(wait_for_server=False)
        client = self.get_client(server.topic)
        for _ in range(3):
            reply = client.call_a()
            self.assertIsNotNone(reply)
        self.assertEqual(3, len(client.replies))



#    Copyright 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import logging
import threading

import futurist
import six

from oslo_messaging._drivers.zmq_driver.client import zmq_response
from oslo_messaging._drivers.zmq_driver import zmq_async
from oslo_messaging._drivers.zmq_driver import zmq_names
from oslo_messaging._drivers.zmq_driver import zmq_version
from oslo_messaging._i18n import _LE

LOG = logging.getLogger(__name__)

zmq = zmq_async.import_zmq()


def suppress_errors(func):
    @six.wraps(func)
    def silent_func(self, socket):
        try:
            return func(self, socket)
        except Exception as e:
            LOG.error(_LE("Receiving message failed: %r"), e)
            # NOTE(gdavoian): drop the left parts of a broken message, since
            # they most likely will lead to additional exceptions
            if socket.getsockopt(zmq.RCVMORE):
                socket.recv_multipart()
    return silent_func


@six.add_metaclass(abc.ABCMeta)
class ReceiverBase(object):
    """Base response receiving interface."""

    def __init__(self, conf):
        self.conf = conf
        self._lock = threading.Lock()
        self._requests = {}
        self._poller = zmq_async.get_poller()
        self._receive_response_versions = \
            zmq_version.get_method_versions(self, 'receive_response')
        self._executor = zmq_async.get_executor(self._run_loop)
        self._executor.execute()

    def register_socket(self, socket):
        """Register a socket for receiving data."""
        self._poller.register(socket, self.receive_response)

    def unregister_socket(self, socket):
        """Unregister a socket from receiving data."""
        self._poller.unregister(socket)

    @abc.abstractmethod
    def receive_response(self, socket):
        """Receive a response (ack or reply) and return it."""

    def track_request(self, request):
        """Track a request via already registered sockets and return
        a pair of ack and reply futures for monitoring all possible
        types of responses for the given request.
        """
        message_id = request.message_id
        futures = self._get_futures(message_id)
        if futures is None:
            ack_future = reply_future = None
            if self.conf.oslo_messaging_zmq.rpc_use_acks:
                ack_future = futurist.Future()
            if request.msg_type == zmq_names.CALL_TYPE:
                reply_future = futurist.Future()
            futures = (ack_future, reply_future)
            self._set_futures(message_id, futures)
        return futures

    def untrack_request(self, request):
        """Untrack a request and stop monitoring any responses."""
        self._pop_futures(request.message_id)

    def stop(self):
        self._poller.close()
        self._executor.stop()

    def _get_futures(self, message_id):
        with self._lock:
            return self._requests.get(message_id)

    def _set_futures(self, message_id, futures):
        with self._lock:
            self._requests[message_id] = futures

    def _pop_futures(self, message_id):
        with self._lock:
            return self._requests.pop(message_id, None)

    def _run_loop(self):
        response, socket = \
            self._poller.poll(self.conf.oslo_messaging_zmq.rpc_poll_timeout)
        if response is None:
            return
        message_type, message_id = response.msg_type, response.message_id
        futures = self._get_futures(message_id)
        if futures is not None:
            ack_future, reply_future = futures
            if message_type == zmq_names.REPLY_TYPE:
                reply_future.set_result(response)
            else:
                ack_future.set_result(response)
            LOG.debug("Received %(msg_type)s for %(msg_id)s",
                      {"msg_type": zmq_names.message_type_str(message_type),
                       "msg_id": message_id})

    def _get_receive_response_version(self, version):
        receive_response_version = self._receive_response_versions.get(version)
        if receive_response_version is None:
            raise zmq_version.UnsupportedMessageVersionError(version)
        return receive_response_version


class ReceiverProxy(ReceiverBase):

    @suppress_errors
    def receive_response(self, socket):
        empty = socket.recv()
        assert empty == b'', "Empty delimiter expected!"
        message_version = socket.recv_string()
        assert message_version != b'', "Valid message version expected!"

        receive_response_version = \
            self._get_receive_response_version(message_version)
        return receive_response_version(socket)

    def _receive_response_v_1_0(self, socket):
        reply_id = socket.recv()
        assert reply_id != b'', "Valid reply id expected!"
        message_type = int(socket.recv())
        assert message_type in zmq_names.RESPONSE_TYPES, "Response expected!"
        message_id = socket.recv_string()
        assert message_id != '', "Valid message id expected!"
        if message_type == zmq_names.REPLY_TYPE:
            reply_body, failure = socket.recv_loaded()
            reply = zmq_response.Reply(message_id=message_id,
                                       reply_id=reply_id,
                                       reply_body=reply_body,
                                       failure=failure)
            return reply
        else:
            ack = zmq_response.Ack(message_id=message_id,
                                   reply_id=reply_id)
            return ack


class ReceiverDirect(ReceiverBase):

    @suppress_errors
    def receive_response(self, socket):
        empty = socket.recv()
        assert empty == b'', "Empty delimiter expected!"
        message_version = socket.recv_string()
        assert message_version != b'', "Valid message version expected!"

        receive_response_version = \
            self._get_receive_response_version(message_version)
        return receive_response_version(socket)

    def _receive_response_v_1_0(self, socket):
        message_type = int(socket.recv())
        assert message_type in zmq_names.RESPONSE_TYPES, "Response expected!"
        message_id = socket.recv_string()
        assert message_id != '', "Valid message id expected!"
        if message_type == zmq_names.REPLY_TYPE:
            reply_body, failure = socket.recv_loaded()
            reply = zmq_response.Reply(message_id=message_id,
                                       reply_body=reply_body,
                                       failure=failure)
            return reply
        else:
            ack = zmq_response.Ack(message_id=message_id)
            return ack



#    Copyright 2015-2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_messaging._drivers.zmq_driver import zmq_async

zmq = zmq_async.import_zmq()


FIELD_MSG_ID = 'message_id'
FIELD_REPLY_ID = 'reply_id'
FIELD_MSG_VERSION = 'message_version'
FIELD_REPLY_BODY = 'reply_body'
FIELD_FAILURE = 'failure'


REPLY_ID_IDX = 0
EMPTY_IDX = 1
MESSAGE_VERSION_IDX = 2
MESSAGE_TYPE_IDX = 3
ROUTING_KEY_IDX = 4
MESSAGE_ID_IDX = 5


DEFAULT_TYPE = 0
CALL_TYPE = 1
CAST_TYPE = 2
CAST_FANOUT_TYPE = 3
NOTIFY_TYPE = 4
REPLY_TYPE = 5
ACK_TYPE = 6

REQUEST_TYPES = (CALL_TYPE,
                 CAST_TYPE,
                 CAST_FANOUT_TYPE,
                 NOTIFY_TYPE)

RESPONSE_TYPES = (REPLY_TYPE, ACK_TYPE)

MESSAGE_TYPES = REQUEST_TYPES + RESPONSE_TYPES

MULTISEND_TYPES = (CAST_FANOUT_TYPE, NOTIFY_TYPE)
DIRECT_TYPES = (CALL_TYPE, CAST_TYPE) + RESPONSE_TYPES
CAST_TYPES = (CAST_TYPE, CAST_FANOUT_TYPE)
NOTIFY_TYPES = (NOTIFY_TYPE,)
NON_BLOCKING_TYPES = CAST_TYPES + NOTIFY_TYPES


def socket_type_str(socket_type):
    zmq_socket_str = {zmq.DEALER: "DEALER",
                      zmq.ROUTER: "ROUTER",
                      zmq.PUSH: "PUSH",
                      zmq.PULL: "PULL",
                      zmq.REQ: "REQ",
                      zmq.REP: "REP",
                      zmq.PUB: "PUB",
                      zmq.SUB: "SUB"}
    return zmq_socket_str[socket_type]


def message_type_str(message_type):
    msg_type_str = {CALL_TYPE: "CALL",
                    CAST_TYPE: "CAST",
                    CAST_FANOUT_TYPE: "CAST_FANOUT",
                    NOTIFY_TYPE: "NOTIFY",
                    REPLY_TYPE: "REPLY",
                    ACK_TYPE: "ACK"}
    return msg_type_str.get(message_type, "UNKNOWN")



#    Copyright 2015-2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import uuid

import six

from oslo_messaging._drivers.zmq_driver.server.consumers \
    import zmq_consumer_base
from oslo_messaging._drivers.zmq_driver.server import zmq_incoming_message
from oslo_messaging._drivers.zmq_driver import zmq_address
from oslo_messaging._drivers.zmq_driver import zmq_async
from oslo_messaging._drivers.zmq_driver import zmq_names
from oslo_messaging._drivers.zmq_driver import zmq_socket
from oslo_messaging._drivers.zmq_driver import zmq_updater
from oslo_messaging._drivers.zmq_driver import zmq_version
from oslo_messaging._i18n import _LE, _LI

LOG = logging.getLogger(__name__)

zmq = zmq_async.import_zmq()


class SubConsumer(zmq_consumer_base.ConsumerBase):

    def __init__(self, conf, poller, server):
        super(SubConsumer, self).__init__(conf, poller, server)
        self.matchmaker = SubscriptionMatchmakerWrapper(conf,
                                                        server.matchmaker)
        self.target = server.target
        self.socket = zmq_socket.ZmqSocket(self.conf, self.context, zmq.SUB,
                                           immediate=False,
                                           identity=self._generate_identity())
        self.sockets.append(self.socket)
        self.host = self.socket.handle.identity
        self._subscribe_to_topic()
        self._receive_request_versions = \
            zmq_version.get_method_versions(self, 'receive_request')
        self.connection_updater = SubscriberConnectionUpdater(
            conf, self.matchmaker, self.socket)
        self.poller.register(self.socket, self.receive_request)
        LOG.info(_LI("[%s] Run SUB consumer"), self.host)

    def _generate_identity(self):
        return six.b(self.conf.oslo_messaging_zmq.rpc_zmq_host + '/') + \
            zmq_address.target_to_subscribe_filter(self.target) + \
            six.b('/' + str(uuid.uuid4()))

    def _subscribe_to_topic(self):
        topic_filter = zmq_address.target_to_subscribe_filter(self.target)
        self.socket.setsockopt(zmq.SUBSCRIBE, topic_filter)
        LOG.debug("[%(host)s] Subscribing to topic %(filter)s",
                  {"host": self.host, "filter": topic_filter})

    def _get_receive_request_version(self, version):
        receive_request_version = self._receive_request_versions.get(version)
        if receive_request_version is None:
            raise zmq_version.UnsupportedMessageVersionError(version)
        return receive_request_version

    def _receive_request_v_1_0(self, topic_filter, socket):
        message_type = int(socket.recv())
        assert message_type in zmq_names.MULTISEND_TYPES, "Fanout expected!"
        message_id = socket.recv()
        context, message = socket.recv_loaded()
        LOG.debug("[%(host)s] Received on topic %(filter)s message %(msg_id)s "
                  "(v%(msg_version)s)",
                  {'host': self.host,
                   'filter': topic_filter,
                   'msg_id': message_id,
                   'msg_version': '1.0'})
        return context, message

    def receive_request(self, socket):
        try:
            topic_filter = socket.recv()
            message_version = socket.recv_string()
            receive_request_version = \
                self._get_receive_request_version(message_version)
            context, message = receive_request_version(topic_filter, socket)
            return zmq_incoming_message.ZmqIncomingMessage(context, message)
        except (zmq.ZMQError, AssertionError, ValueError,
                zmq_version.UnsupportedMessageVersionError) as e:
            LOG.error(_LE("Receiving message failed: %s"), str(e))
            # NOTE(gdavoian): drop the left parts of a broken message
            if socket.getsockopt(zmq.RCVMORE):
                socket.recv_multipart()

    def cleanup(self):
        LOG.info(_LI("[%s] Destroy SUB consumer"), self.host)
        self.connection_updater.cleanup()
        super(SubConsumer, self).cleanup()


class SubscriptionMatchmakerWrapper(object):

    def __init__(self, conf, matchmaker):
        self.conf = conf
        self.matchmaker = matchmaker

    def get_publishers(self):
        conf_publishers = self.conf.oslo_messaging_zmq.subscribe_on
        LOG.debug("Publishers taken from configuration %s", conf_publishers)
        if conf_publishers:
            return [(publisher, None) for publisher in conf_publishers]
        return self.matchmaker.get_publishers()


class SubscriberConnectionUpdater(zmq_updater.ConnectionUpdater):

    def _update_connection(self):
        publishers = self.matchmaker.get_publishers()
        for publisher_address, router_address in publishers:
            self.socket.connect_to_host(publisher_address)
        LOG.debug("[%s] SUB consumer connected to publishers %s",
                  self.socket.handle.identity, publishers)



#    Copyright 2015-2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

from oslo_messaging._drivers.zmq_driver.proxy.central \
    import zmq_publisher_proxy
from oslo_messaging._drivers.zmq_driver.proxy \
    import zmq_base_proxy
from oslo_messaging._drivers.zmq_driver.proxy import zmq_sender
from oslo_messaging._drivers.zmq_driver import zmq_async
from oslo_messaging._drivers.zmq_driver import zmq_names
from oslo_messaging._drivers.zmq_driver import zmq_updater
from oslo_messaging._i18n import _LI

LOG = logging.getLogger(__name__)

zmq = zmq_async.import_zmq()


class SingleRouterProxy(zmq_base_proxy.ProxyBase):

    PROXY_TYPE = "ROUTER"

    def __init__(self, conf, context, matchmaker):
        super(SingleRouterProxy, self).__init__(conf, context, matchmaker)

        port = conf.zmq_proxy_opts.frontend_port
        self.fe_router_socket = zmq_base_proxy.create_socket(
            conf, context, port, zmq.ROUTER)

        self.poller.register(self.fe_router_socket, self._receive_message)

        self.publisher = zmq_publisher_proxy.PublisherProxy(conf, matchmaker)

        self.router_sender = zmq_sender.CentralRouterSender()
        self.ack_sender = zmq_sender.CentralAckSender()

        self._router_updater = self._create_router_updater()

    def run(self):
        message, socket = self.poller.poll()
        if message is None:
            return

        message_type = int(message[zmq_names.MESSAGE_TYPE_IDX])
        if self.conf.oslo_messaging_zmq.use_pub_sub and \
                message_type in zmq_names.MULTISEND_TYPES:
            self.publisher.send_request(message)
            if socket is self.fe_router_socket and \
                    self.conf.zmq_proxy_opts.ack_pub_sub:
                self.ack_sender.send_message(socket, message)
        else:
            self.router_sender.send_message(
                self._get_socket_to_dispatch_on(socket), message)

    def _create_router_updater(self):
        return RouterUpdater(
            self.conf, self.matchmaker, self.publisher.host,
            self.fe_router_socket.connect_address,
            self.fe_router_socket.connect_address)

    def _get_socket_to_dispatch_on(self, socket):
        return self.fe_router_socket

    def cleanup(self):
        super(SingleRouterProxy, self).cleanup()
        self._router_updater.cleanup()
        self.fe_router_socket.close()
        self.publisher.cleanup()


class DoubleRouterProxy(SingleRouterProxy):

    PROXY_TYPE = "ROUTER-ROUTER"

    def __init__(self, conf, context, matchmaker):
        port = conf.zmq_proxy_opts.backend_port
        self.be_router_socket = zmq_base_proxy.create_socket(
            conf, context, port, zmq.ROUTER)
        super(DoubleRouterProxy, self).__init__(conf, context, matchmaker)
        self.poller.register(self.be_router_socket, self._receive_message)

    def _create_router_updater(self):
        return RouterUpdater(
            self.conf, self.matchmaker, self.publisher.host,
            self.fe_router_socket.connect_address,
            self.be_router_socket.connect_address)

    def _get_socket_to_dispatch_on(self, socket):
        return self.be_router_socket \
            if socket is self.fe_router_socket \
            else self.fe_router_socket

    def cleanup(self):
        super(DoubleRouterProxy, self).cleanup()
        self.be_router_socket.close()


class RouterUpdater(zmq_updater.UpdaterBase):
    """This entity performs periodic async updates
    from router proxy to the matchmaker.
    """

    def __init__(self, conf, matchmaker, publisher_address, fe_router_address,
                 be_router_address):
        self.publisher_address = publisher_address
        self.fe_router_address = fe_router_address
        self.be_router_address = be_router_address
        super(RouterUpdater, self).__init__(
            conf, matchmaker, self._update_records,
            conf.oslo_messaging_zmq.zmq_target_update)

    def _update_records(self):
        self.matchmaker.register_publisher(
            (self.publisher_address, self.fe_router_address),
            expire=self.conf.oslo_messaging_zmq.zmq_target_expire)
        LOG.info(_LI("[PUB:%(pub)s, ROUTER:%(router)s] Update PUB publisher"),
                 {"pub": self.publisher_address,
                  "router": self.fe_router_address})
        self.matchmaker.register_router(
            self.be_router_address,
            expire=self.conf.oslo_messaging_zmq.zmq_target_expire)
        LOG.info(_LI("[Backend ROUTER:%(router)s] Update ROUTER"),
                 {"router": self.be_router_address})

    def cleanup(self):
        super(RouterUpdater, self).cleanup()
        self.matchmaker.unregister_publisher(
            (self.publisher_address, self.fe_router_address))
        self.matchmaker.unregister_router(
            self.be_router_address)



#    Copyright 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import logging
import random
import socket
import threading
import time

from oslo_config import cfg
import pika
from pika import credentials as pika_credentials

from oslo_messaging._drivers.pika_driver import pika_commons as pika_drv_cmns
from oslo_messaging._drivers.pika_driver import pika_connection
from oslo_messaging._drivers.pika_driver import pika_exceptions as pika_drv_exc

LOG = logging.getLogger(__name__)

# constant for setting tcp_user_timeout socket option
# (it should be defined in 'select' module of standard library in future)
TCP_USER_TIMEOUT = 18

# constants for creating connection statistics
HOST_CONNECTION_LAST_TRY_TIME = "last_try_time"
HOST_CONNECTION_LAST_SUCCESS_TRY_TIME = "last_success_try_time"

pika_opts = [
    cfg.IntOpt('channel_max',
               help='Maximum number of channels to allow'),
    cfg.IntOpt('frame_max',
               help='The maximum byte size for an AMQP frame'),
    cfg.IntOpt('heartbeat_interval', default=3,
               help="How often to send heartbeats for consumer's connections"),
    cfg.BoolOpt('ssl',
                help='Enable SSL'),
    cfg.DictOpt('ssl_options',
                help='Arguments passed to ssl.wrap_socket'),
    cfg.FloatOpt('socket_timeout', default=0.25,
                 help="Set socket timeout in seconds for connection's socket"),
    cfg.FloatOpt('tcp_user_timeout', default=0.25,
                 help="Set TCP_USER_TIMEOUT in seconds for connection's "
                      "socket"),
    cfg.FloatOpt('host_connection_reconnect_delay', default=0.25,
                 help="Set delay for reconnection to some host which has "
                      "connection error"),
    cfg.StrOpt('connection_factory', default="single",
               choices=["new", "single", "read_write"],
               help='Connection factory implementation')
]


class PikaConnectionFactory(object):

    def __init__(self, url, conf):
        self._url = url
        self._conf = conf

        self._connection_lock = threading.RLock()

        if not url.hosts:
            raise ValueError("You should provide at least one RabbitMQ host")

        # initializing connection parameters for configured RabbitMQ hosts
        self._common_pika_params = {
            'virtual_host': url.virtual_host,
            'channel_max': conf.oslo_messaging_pika.channel_max,
            'frame_max': conf.oslo_messaging_pika.frame_max,
            'ssl': conf.oslo_messaging_pika.ssl,
            'ssl_options': conf.oslo_messaging_pika.ssl_options,
            'socket_timeout': conf.oslo_messaging_pika.socket_timeout
        }

        self._host_list = url.hosts
        self._heartbeat_interval = conf.oslo_messaging_pika.heartbeat_interval
        self._host_connection_reconnect_delay = (
            conf.oslo_messaging_pika.host_connection_reconnect_delay
        )
        self._tcp_user_timeout = conf.oslo_messaging_pika.tcp_user_timeout

        self._connection_host_status = {}

        self._cur_connection_host_num = random.randint(
            0, len(url.hosts) - 1
        )

    def cleanup(self):
        pass

    def create_connection(self, for_listening=False):
        """Create and return connection to any available host.

        :return: created connection
        :raise: ConnectionException if all hosts are not reachable
        """

        with self._connection_lock:

            host_count = len(self._host_list)
            connection_attempts = host_count

            while connection_attempts > 0:
                self._cur_connection_host_num += 1
                self._cur_connection_host_num %= host_count
                try:
                    return self._create_host_connection(
                        self._cur_connection_host_num, for_listening
                    )
                except pika_drv_cmns.PIKA_CONNECTIVITY_ERRORS as e:
                    LOG.warning("Can't establish connection to host. %s", e)
                except pika_drv_exc.HostConnectionNotAllowedException as e:
                    LOG.warning("Connection to host is not allowed. %s", e)

                connection_attempts -= 1

            raise pika_drv_exc.EstablishConnectionException(
                "Can not establish connection to any configured RabbitMQ "
                "host: " + str(self._host_list)
            )

    def _set_tcp_user_timeout(self, s):
        if not self._tcp_user_timeout:
            return
        try:
            s.setsockopt(
                socket.IPPROTO_TCP, TCP_USER_TIMEOUT,
                int(self._tcp_user_timeout * 1000)
            )
        except socket.error:
            LOG.warning(
                "Whoops, this kernel doesn't seem to support TCP_USER_TIMEOUT."
            )

    def _create_host_connection(self, host_index, for_listening):
        """Create new connection to host #host_index

        :param host_index: Integer, number of host for connection establishing
        :param for_listening: Boolean, creates connection for listening
            if True
        :return: New connection
        """
        host = self._host_list[host_index]

        cur_time = time.time()

        host_connection_status = self._connection_host_status.get(host)

        if host_connection_status is None:
            host_connection_status = {
                HOST_CONNECTION_LAST_SUCCESS_TRY_TIME: 0,
                HOST_CONNECTION_LAST_TRY_TIME: 0
            }
            self._connection_host_status[host] = host_connection_status

        last_success_time = host_connection_status[
            HOST_CONNECTION_LAST_SUCCESS_TRY_TIME
        ]
        last_time = host_connection_status[
            HOST_CONNECTION_LAST_TRY_TIME
        ]

        # raise HostConnectionNotAllowedException if we tried to establish
        # connection in last 'host_connection_reconnect_delay' and got
        # failure
        if (last_time != last_success_time and
                cur_time - last_time <
                self._host_connection_reconnect_delay):
            raise pika_drv_exc.HostConnectionNotAllowedException(
                "Connection to host #{} is not allowed now because of "
                "previous failure".format(host_index)
            )

        try:
            connection = self._do_create_host_connection(
                host, for_listening
            )
            self._connection_host_status[host][
                HOST_CONNECTION_LAST_SUCCESS_TRY_TIME
            ] = cur_time

            return connection
        finally:
            self._connection_host_status[host][
                HOST_CONNECTION_LAST_TRY_TIME
            ] = cur_time

    def _do_create_host_connection(self, host, for_listening):
        connection_params = pika.ConnectionParameters(
            host=host.hostname,
            port=host.port,
            credentials=pika_credentials.PlainCredentials(
                host.username, host.password
            ),
            heartbeat_interval=(
                self._heartbeat_interval if for_listening else None
            ),
            **self._common_pika_params
        )
        if for_listening:
            connection = pika_connection.ThreadSafePikaConnection(
                parameters=connection_params
            )
        else:
            connection = pika.BlockingConnection(
                parameters=connection_params
            )
            connection.params = connection_params

        self._set_tcp_user_timeout(connection._impl.socket)
        return connection


class NotClosableConnection(object):
    def __init__(self, connection):
        self._connection = connection

    def __getattr__(self, item):
        return getattr(self._connection, item)

    def close(self):
        pass


class SinglePikaConnectionFactory(PikaConnectionFactory):
    def __init__(self, url, conf):
        super(SinglePikaConnectionFactory, self).__init__(url, conf)
        self._connection = None

    def create_connection(self, for_listening=False):
        with self._connection_lock:
            if self._connection is None or not self._connection.is_open:
                self._connection = (
                    super(SinglePikaConnectionFactory, self).create_connection(
                        True
                    )
                )
            return NotClosableConnection(self._connection)

    def cleanup(self):
        with self._connection_lock:
            if self._connection is not None and self._connection.is_open:
                try:
                    self._connection.close()
                except Exception:
                    LOG.warning(
                        "Unexpected exception during connection closing",
                        exc_info=True
                    )
            self._connection = None


class ReadWritePikaConnectionFactory(PikaConnectionFactory):
    def __init__(self, url, conf):
        super(ReadWritePikaConnectionFactory, self).__init__(url, conf)
        self._read_connection = None
        self._write_connection = None

    def create_connection(self, for_listening=False):
        with self._connection_lock:
            if for_listening:
                if (self._read_connection is None or
                        not self._read_connection.is_open):
                    self._read_connection = super(
                        ReadWritePikaConnectionFactory, self
                    ).create_connection(True)
                return NotClosableConnection(self._read_connection)
            else:
                if (self._write_connection is None or
                        not self._write_connection.is_open):
                    self._write_connection = super(
                        ReadWritePikaConnectionFactory, self
                    ).create_connection(True)
                return NotClosableConnection(self._write_connection)

    def cleanup(self):
        with self._connection_lock:
            if (self._read_connection is not None and
                    self._read_connection.is_open):
                try:
                    self._read_connection.close()
                except Exception:
                    LOG.warning(
                        "Unexpected exception during connection closing",
                        exc_info=True
                    )
            self._read_connection = None

            if (self._write_connection is not None and
                    self._write_connection.is_open):
                try:
                    self._write_connection.close()
                except Exception:
                    LOG.warning(
                        "Unexpected exception during connection closing",
                        exc_info=True
                    )
            self._write_connection = None



#    Copyright 2015-2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

import six

from oslo_messaging._drivers.zmq_driver import zmq_names


@six.add_metaclass(abc.ABCMeta)
class Response(object):

    def __init__(self, message_id=None, reply_id=None, message_version=None):
        if self.msg_type not in zmq_names.RESPONSE_TYPES:
            raise RuntimeError("Unknown response type!")

        self._message_id = message_id
        self._reply_id = reply_id
        self._message_version = message_version

    @abc.abstractproperty
    def msg_type(self):
        """ZMQ response type"""

    @property
    def message_id(self):
        return self._message_id

    @property
    def reply_id(self):
        return self._reply_id

    @property
    def message_version(self):
        return self._message_version

    def to_dict(self):
        return {zmq_names.FIELD_MSG_ID: self._message_id,
                zmq_names.FIELD_REPLY_ID: self._reply_id,
                zmq_names.FIELD_MSG_VERSION: self._message_version}

    def __str__(self):
        return str(self.to_dict())


class Ack(Response):

    msg_type = zmq_names.ACK_TYPE


class Reply(Response):

    msg_type = zmq_names.REPLY_TYPE

    def __init__(self, message_id=None, reply_id=None, message_version=None,
                 reply_body=None, failure=None):
        super(Reply, self).__init__(message_id, reply_id, message_version)
        self._reply_body = reply_body
        self._failure = failure

    @property
    def reply_body(self):
        return self._reply_body

    @property
    def failure(self):
        return self._failure

    def to_dict(self):
        dict_ = super(Reply, self).to_dict()
        dict_.update({zmq_names.FIELD_REPLY_BODY: self._reply_body,
                      zmq_names.FIELD_FAILURE: self._failure})
        return dict_
